#include "c_enemy.h"
#include "c_game.h"
#include "c_sprite.h"
#include <math.h>

c_enemy::c_enemy()
{
	m_state        = ES_SPAWNING;
	m_prev_state   = ES_SPAWNING;
	m_bullet_timer = 0;
}

c_enemy::c_enemy(c_game *game, string dat_file):c_character_actor(game, dat_file)
{
	m_state         = ES_SPAWNING;
	m_bullet_timer  = 0;
	m_kill_score    = 0;
	
	// Randomise speed
	m_speed += (rand() % 40) / 20.0;
}

c_enemy::~c_enemy()
{
}

void c_enemy::update(void)
{
	// Calculate vector from enemy to player
	float dx = (m_x + ((float)(m_sprite->get_tile_width()) / 2.0)) - (m_p_game->get_player()->get_x() + ((float)(m_p_game->get_player()->get_sprite()->get_tile_width()) / 2.0));
	float dy = (m_y + ((float)(m_sprite->get_tile_height()) / 2.0)) - (m_p_game->get_player()->get_y() + ((float)(m_p_game->get_player()->get_sprite()->get_tile_height()) / 2.0));
	
	float mag = 0.0;
	float proj_x = 0.0, proj_y = 0.0;
	
	// AI: set state based on health (chasing or fleeing)
	if (m_health < 30.0 && m_state == ES_CHASING)
		set_state(ES_FLEEING);
	if (m_health > 30.0 && m_state == ES_FLEEING)
		set_state(ES_CHASING);
	
	if (m_state == ES_SPAWNING)
	{
		set_anim("SPWN");
		
		// Check if anim is complete, if so set state to ES_CHASING
		if (m_sprite->is_curr_anim_finished())
			set_state(ES_CHASING);
		
		c_character_actor::update();
		
		// Enemy cannot move while spawning
		m_xvel = 0.0;
		m_yvel = 0.0;
		
		return;
	}
	
	if (m_state == ES_DYING)
	{
		set_anim("DYIG");
		
		// Check if anim is complete, if so set state to ES_DEAD
		if (m_sprite->is_curr_anim_finished())
			set_state(ES_DEAD);
		
		c_character_actor::update();
		
		// Enemy cannot move whilst dying
		m_xvel = 0.0;
		m_yvel = 0.0;
		
		return;
	}
	
	if (m_state == ES_DEAD)
	{
		// Nothing to do here, move along...
		return;
	}
	
	if (m_state == ES_ATTACK_GUN)
	{
		// If enemy has finished firing, return to the previous state
		if (m_sprite->is_curr_anim_finished())
			set_state(m_prev_state);
		
		c_character_actor::update();
		
		// Enemy cannot move whilst firing
		m_xvel = 0.0;
		m_yvel = 0.0;
		
		return;
	}
	
	if (m_state == ES_ATTACK_MELEE)
	{
		// Currently unused
		c_character_actor::update();
		m_xvel = 0.0;
		m_yvel = 0.0;
		return;
	}
	
	mag = sqrt(dx * dx + dy * dy);
	
	if (mag > 0)
	{
		// Normalise the enemy->player vector and multiply by m_speed to get velocity vector
		m_xvel = abs((dx / mag) * m_speed);
		m_yvel = abs((dy / mag) * m_speed);
	}
	else
	{
		m_xvel = 0.0;
		m_yvel = 0.0;
	}
	
	if (m_state == ES_CHASING)
	{
		// Make sure enemy moves towards the player
		if (m_p_game->get_player()->get_x() < m_x)
			m_xvel = -m_xvel;
		if (m_p_game->get_player()->get_y() < m_y)
			m_yvel = -m_yvel;
	}
	else if (m_state == ES_FLEEING)
	{
		// Recharge the enemy's health
		if (m_health < 75.0)
			m_health += 0.5;
		
		// Make sure the enemy moves away from the player
		if (m_p_game->get_player()->get_x() > m_x)
			m_xvel = -m_xvel;
		if (m_p_game->get_player()->get_y() > m_y)
			m_yvel = -m_yvel;
	}
	
	// Set an appropriate animation based on velocity
	if (m_xvel < 0.0)
		set_anim("MVLT");
	else if (m_xvel > 0.0)
		set_anim("MVRT");
	else if (m_yvel < 0.0)
		set_anim("MVUP");
	else if (m_yvel > 0.0)
		set_anim("MVDN");
	else
		set_anim("IDLE");
	
	// Decrement bullet timer for rate limiting
	m_bullet_timer--;
	
	// Determine randomly if enemy is going to fire a projectile
	if (m_bullet_timer <= 0)
	{
		m_bullet_timer = 0;
		if (rand() % 100 == 0)
		{
			// Fire projectile towards player
			proj_x = -(dx / mag) * ENEMY_PROJECTILE_SPEED;
			proj_y = -(dy / mag) * ENEMY_PROJECTILE_SPEED;
			fire_projectile(proj_x, proj_y);
		}
	}
	
	c_character_actor::update();
	
	// Make sure enemy doesn't move out of the game bounds
	if (m_x < m_p_game->get_bound_left())
		m_x = m_p_game->get_bound_left();
	if (m_x + m_sprite->get_tile_width() > m_p_game->get_bound_right())
		m_x = m_p_game->get_bound_right() - m_sprite->get_tile_width();
	if (m_y < m_p_game->get_bound_top())
		m_y = m_p_game->get_bound_top();
	if (m_y + m_sprite->get_tile_height() > m_p_game->get_bound_bottom())
		m_y = m_p_game->get_bound_bottom() - m_sprite->get_tile_height();
}

void c_enemy::set_state(ENEMY_STATE state)
{
	m_prev_state = m_state;
	m_state = state;
}

ENEMY_STATE c_enemy::get_state(void)
{
	return m_state;
}

void c_enemy::fire_projectile(float vx, float vy)
{
	int x = 0, y = 0;
	
	// Start projectile from the enemy's centre point
	x = m_x + ((float)m_sprite->get_tile_width() / 2.0);
	y = m_y + ((float)m_sprite->get_tile_height() / 2.0);
	
	// Use c_game to fire the projectile
	m_p_game->fire_projectile(this, DT_PLAYER, x, y, vx, vy, "data/Enemy_bullet.dat");
	
	// Set the timer to limit enemy firing rate
	m_bullet_timer = ENEMY_PROJECTILE_LIMITER;
	
	set_state(ES_ATTACK_GUN);
}

void c_enemy::set_kill_score(int score)
{
	m_kill_score = score;
}

int c_enemy::get_kill_score(void)
{
	return m_kill_score;
}
