#include "c_game.h"
#include "c_actor.h"
#include "c_sprite.h"

// Include platform-specific headers
#ifdef LINUX
	#include <sys/time.h>
	#include <inttypes.h>
	#include <unistd.h>
#else
	#include <winalleg.h>
#endif

c_game::c_game()
{
	m_state              = GS_NONE;
	m_run_state          = GRS_MAINMENU;
	m_buf                = NULL;
	m_backdrop           = NULL;
	m_game_over_backdrop = NULL;
	m_mouse_cursor       = NULL;
	m_frame              = 0;
	m_p_bitmap_factory   = new c_bitmap_factory;
	m_player             = NULL;
	m_p_enemies          = NULL;
	m_p_projectiles      = NULL;
	m_p_particle_systems = NULL;
	m_p_powerup          = NULL;
	m_p_sfx              = NULL;
	m_p_jukebox          = NULL;
	m_p_menu_main        = NULL;
	m_curr_main_track    = JBT_MAIN_1;
	m_enemy_count        = 0;
	m_shots              = 0;
	m_shots_hit          = 0;
	m_score              = 0;
	m_bound_left         = 0;
	m_bound_right        = 0;
	m_bound_top          = 0;
	m_bound_bottom       = 0;
	allegro_init();
}

c_game::~c_game()
{
	delete m_p_bitmap_factory;
}

c_bitmap_factory* c_game::get_bitmap_factory(void)
{
	return m_p_bitmap_factory;
}

c_player* c_game::get_player(void)
{
	return m_player;
}

void c_game::sleep(int ms)
{
#ifdef LINUX
	usleep(ms * 1000); // Linux-specific thread sleep function
#else
	Sleep(ms); // Windows-specific thread sleep function
#endif
}

bool c_game::key_press(int key_id)
{
	if (key[key_id])
	{
		// If key is pressed, sleep until it becomes unpressed then return true
		while (key[key_id])
			sleep(5);
		return true;
	}
	return false;
}

void c_game::keyboard_handler(void)
{
	// Keyboard should be handled differently at different run states
	switch (m_run_state)
	{
		case GRS_MAINMENU:
			// Menu object will handle all keyboard events
			break;
		case GRS_PLAYING: // Pass through to GRS_PAUSE as key handling is the same for both
		case GRS_PAUSE:
			if (key_press(KEY_M)) // Cycle track 1, 2, no music
			{
				if (m_curr_main_track == JBT_MAIN_1)
					m_curr_main_track = JBT_MAIN_2;
				else if (m_curr_main_track == JBT_MAIN_2)
					m_curr_main_track = JBT_NONE;
				else
					m_curr_main_track = JBT_MAIN_1;
				m_p_jukebox->play_track(m_curr_main_track, true);
			}
			if (key_press(KEY_N)) // Toggle sound effects
				m_p_sfx->set_enabled(!m_p_sfx->is_enabled());
			if (key_press(KEY_P)) // Toggle paused/unpaused
				set_run_state(m_run_state == GRS_PAUSE ? GRS_PLAYING : GRS_PAUSE);
			if (key_press(KEY_ESC)) // Player wants to quit
				set_run_state(GRS_GAMEOVER);
			break;
		case GRS_GAMEOVER:
			if (key_press(KEY_ENTER)) // Player wants to play again
			{
				if (reset())
					set_run_state(GRS_PLAYING);
				else
				{
					allegro_message("ERROR: Unable to reset game, aborting...", "");
					m_state = GS_QUIT;
				}
			}
			else if (key_press(KEY_ESC)) // Player wants to return to the mani menu
			{
				reset();
				set_run_state(GRS_MAINMENU);
			}
			break;
	}
}

void c_game::menu_handler(c_menu *menu, int item_id)
{
	if (menu == m_p_menu_main)
	{
		// This is the main menu calling!
		switch (item_id)
		{
			case 0: // Play game
				set_run_state(GRS_PLAYING);
				break;
			case 1: // Quit
				m_state = GS_QUIT;
				break;
		}
	}
}

void c_game::add_enemy(int x, int y, float health, int kill_score, string dat_file)
{
	enemy_list *p_node = new enemy_list;
	
	p_node->enemy = new c_enemy(this, dat_file);
	if (p_node->enemy == NULL)
		return;
	
	p_node->enemy->set_x(x);
	p_node->enemy->set_y(y);
	p_node->enemy->set_anim("IDLE");
	p_node->enemy->set_health(health);
	p_node->enemy->set_kill_score(kill_score);
	
	// Add new enemy to the head of the list
	p_node->p_next = m_p_enemies;
	m_p_enemies = p_node;
	
	m_enemy_count++;
}

void c_game::remove_enemy(c_enemy *p_enemy)
{
	enemy_list *p_node = m_p_enemies, *p_prev = NULL;
	
	// Search for enemy in the enemy list
	while (p_node != NULL)
	{
		if (p_node->enemy == p_enemy)
			break;
		p_prev = p_node;
		p_node = p_node->p_next;
	}
	
	// If the enemy was found, remove it from the list and destroy it
	if (p_node != NULL)
	{
		// Blit the enemy to the backdrop first
		p_enemy->draw(m_backdrop);
		
		// If p_node was not the head, set the previous node's p_next pointer
		// to this node's p_next pointer
		// Else set the m_p_enemies pointer to p_node->p_next
		if (p_prev != NULL)
			p_prev->p_next = p_node->p_next;
		else
			m_p_enemies = p_node->p_next;
		
		delete p_node->enemy;
		delete p_node;
		
		m_enemy_count--;
	}
}

void c_game::update_enemies(void)
{
	enemy_list *p_node = m_p_enemies, *p_prev = NULL;

	// Loop through enemy list and update each element
	while (p_node != NULL)
	{
		p_node->enemy->update();
		
		// If enemy is dead, it needs to be removed
		if (p_node->enemy->get_state() == ES_DEAD)
		{
			remove_enemy(p_node->enemy);
			if (p_prev != NULL)
				p_node = p_prev->p_next;
			else
				p_node = m_p_enemies;			
		}
		else
		{
			p_prev = p_node;
			p_node = p_node->p_next;
		}
	}
}

void c_game::draw_enemies(void)
{
	enemy_list *p_node = m_p_enemies;

	// Loop through enemy list and draw each one
	while (p_node != NULL)
	{
		p_node->enemy->draw(m_buf);
		p_node = p_node->p_next;
	}
}

void c_game::fire_projectile(c_character_actor *p_owner, DAMAGE_TYPE damage_type, int x, int y, float x_vel, float y_vel, string dat_file)
{
	bool f_found = false;
	projectile_list *p_node = m_p_projectiles;
	
	if (p_owner == m_player)
	{
		// Increment player shots fired counter
		m_shots++;

		// Play a SFX_PLAYER_FIRE sfx
		m_p_sfx->play_sfx(SFX_PLAYER_FIRE);
	}
	else
	{
		// Play a SFX_ENEMY_FIRE sfx
		m_p_sfx->play_sfx(SFX_ENEMY_FIRE);
	}
	
	// First see if there is a suitable projectile in the list that is inactive
	while (p_node != NULL)
	{
		if (p_node->projectile != NULL)
		{
			if (!p_node->projectile->is_active() && p_node->projectile->get_dat_file_name() == dat_file)
			{
				// We've found a suitable projectile (inactive and of the correct type)
				f_found = true;
				break;
			}
		}
		p_node = p_node->p_next;
	}
	
	// If no suitable projectile was found, spawn a new one
	if (!f_found)
	{
		p_node = new projectile_list;
		if (p_node == NULL)
			return;
		p_node->projectile = new c_projectile(this, dat_file);
		if (p_node->projectile == NULL)
			return;
		p_node->p_next = m_p_projectiles;
		m_p_projectiles = p_node;
	}
	
	// Initialise the projectile and activate it
	p_node->projectile->set_x(x);
	p_node->projectile->set_y(y);
	p_node->projectile->set_xvel(x_vel);
	p_node->projectile->set_yvel(y_vel);
	p_node->projectile->set_damage_type(damage_type);
	p_node->projectile->set_damage_power(25.0);
	p_node->projectile->set_state(PS_NORMAL);
	p_node->projectile->activate();
}

void c_game::update_projectiles(void)
{
	projectile_list *p_node = m_p_projectiles;

	// Loop through projectile list, update them and perform hit testing/handling
	while (p_node != NULL)
	{
		p_node->projectile->update();
		projectile_hit_test(p_node->projectile);
		p_node = p_node->p_next;
	}
}

void c_game::draw_projectiles(void)
{
	projectile_list *p_node = m_p_projectiles;

	// Loop through projectile list and draw them
	while (p_node != NULL)
	{
		p_node->projectile->draw(m_buf);
		p_node = p_node->p_next;
	}
}

void c_game::projectile_hit_test(c_projectile *projectile)
{
	enemy_list *p_enemy = m_p_enemies;
	float px1 = 0.0, py1 = 0.0, px2 = 0.0, py2 = 0.0;
	float  x1 = 0.0,  y1 = 0.0,  x2 = 0.0,  y2 = 0.0;
	float pt_x = 0.0, pt_y = 0.0, pt_vec_x = 0.0, pt_vec_y = 0.0;
	
	if (!projectile->is_active() || projectile->get_state() != PS_NORMAL)
		return;
	
	// Get projectile bounding box
	px1 = projectile->get_x();
	py1 = projectile->get_y();
	px2 = px1 + projectile->get_sprite()->get_tile_width();
	py2 = py1 + projectile->get_sprite()->get_tile_height();
	
	// Enemy hit tests
	if (projectile->get_damage_type() == DT_ENEMY || projectile->get_damage_type() == DT_BOTH)
	{
		// Loop through enemy list
		while (p_enemy != NULL)
		{
			// Get enemy bounding box
			x1 = p_enemy->enemy->get_x();
			y1 = p_enemy->enemy->get_y();
			x2 = x1 + p_enemy->enemy->get_sprite()->get_tile_width();
			y2 = y1 + p_enemy->enemy->get_sprite()->get_tile_height();
			
			if (rect_hit_test(px1, py1, px2, py2, x1, y1, x2, y2))
			{
				// Rectangles intersect, so projectile has hit this enemy
				m_score += SCORE_PER_SHOT_HIT;
				
				if (p_enemy->enemy->hit(projectile->get_damage_power()))
				{
					// c_enemy::hit() returnd true, so enemy is dead
					
					if (p_enemy->enemy->get_state() != ES_DYING)
					{
						// Play a SFX_ENEMY_DIE sfx
						m_p_sfx->play_sfx(SFX_ENEMY_DIE);
						
						p_enemy->enemy->set_state(ES_DYING);
						
						// Increment player's score by enemy's kill score
						m_score += p_enemy->enemy->get_kill_score();					
					}
				}
				projectile->deactivate();
				
				// Spawn a blood particle system, velocity mathing the projectile velocity
				pt_x = projectile->get_x();
				pt_y = projectile->get_y();
				pt_vec_x = projectile->get_xvel();
				pt_vec_y = projectile->get_yvel();
				add_particle_system(PST_BLOOD, pt_x, pt_y, pt_vec_x, pt_vec_y);
				
				// Play SFX_ENEMY_HIT sfx
				m_p_sfx->play_sfx(SFX_ENEMY_HIT);
				
				// Increment player shot hit counter
				m_shots_hit++;
				
				break;
			}
			
			p_enemy = p_enemy->p_next;
		}
	}
	
	// Player hit tests
	if (projectile->get_damage_type() == DT_PLAYER || projectile->get_damage_type() == DT_BOTH)
	{
		// Get player bounding box
		x1 = m_player->get_x();
		y1 = m_player->get_y();
		x2 = x1 + m_player->get_sprite()->get_tile_width();
		y2 = y1 + m_player->get_sprite()->get_tile_height();
		
		if (rect_hit_test(px1, py1, px2, py2, x1, y1, x2, y2))
		{
			// Player has been hit
			m_player->hit(projectile->get_damage_power());
			
			projectile->deactivate();
			
			// Spawn a blood particle system, velocity matching the projectile's velocity
			pt_x = projectile->get_x();
			pt_y = projectile->get_y();
			pt_vec_x = projectile->get_xvel();
			pt_vec_y = projectile->get_yvel();
			add_particle_system(PST_BLOOD, pt_x, pt_y, pt_vec_x, pt_vec_y);
			
			// Play SFX_PLAYER_HIT sfx
			m_p_sfx->play_sfx(SFX_PLAYER_HIT);
			
			if (m_player->get_health() <= 0)
			{
				// Oh no! Player is dead!
				if (m_player->get_state() != PLS_DYING)
				{
					m_player->set_state(PLS_DYING);
					
					// Play a SFX_PLAYER_DIE sfx
					m_p_sfx->play_sfx(SFX_PLAYER_DIE);
				}
			}
		}
	}
	
	// Boundary hit test
	if (px2 <= m_bound_left || px1 >= m_bound_right || py2 <= m_bound_top || py1 >= m_bound_bottom)
	{
		// Projectile has hit a boundary wall
		
		projectile->deactivate();
		
		// Spawn a PST_WALL_DEBRIS particle system, velocity perpendicular to the wall
		if (px2 <= m_bound_left)
		{
			pt_x = px2;
			pt_y = py1;
			pt_vec_x = 1.0;
			pt_vec_y = 0.0;
		}
		else if (px1 >= m_bound_right)
		{
			pt_x = px1;
			pt_y = py1;
			pt_vec_x = -1.0;
			pt_vec_y =  0.0;
		}
		else if (py2 <= m_bound_top)
		{
			pt_x = px1;
			pt_y = py2;
			pt_vec_x = 0.0;
			pt_vec_y = 1.0;
		}
		else if (py1 >= m_bound_bottom)
		{
			pt_x = px1;
			pt_y = py1;
			pt_vec_x =  0.0;
			pt_vec_y = -1.0;
		}
		add_particle_system(PST_WALL_DEBRIS, pt_x, pt_y, pt_vec_x, pt_vec_y);
		
		// Play a SFX_WALL_HIT sfx
		m_p_sfx->play_sfx(SFX_WALL_HIT);
	}
}

void c_game::update_powerup(void)
{
	if (m_p_powerup != NULL)
	{
		m_p_powerup->update();
		if (rect_hit_test(m_player->get_x(), m_player->get_y(), m_player->get_x() + m_player->get_sprite()->get_tile_width(), m_player->get_y() + m_player->get_sprite()->get_tile_height(), m_p_powerup->get_x(), m_p_powerup->get_y(), m_p_powerup->get_x() + m_p_powerup->get_sprite()->get_tile_width(), m_p_powerup->get_y() + m_p_powerup->get_sprite()->get_tile_height()))
		{
			// Player has touched the powerup
			switch (m_p_powerup->get_type())
			{
				case PT_HEALTH:
					// Increase the player's health by the powerup's value
					m_player->set_health(m_player->get_health() + m_p_powerup->get_value());
					
					// Make sure player's health does not go over 100%
					if (m_player->get_health() > PLAYER_HEALTH_MAX)
						m_player->set_health(PLAYER_HEALTH_MAX);
					
					m_p_powerup->deactivate();
					
					// Play SFX_HEALTH_PICKUP sfx
					m_p_sfx->play_sfx(SFX_HEALTH_PICKUP);
					
					break;
			}
		}
		if (!m_p_powerup->is_active())
		{
			delete m_p_powerup;
			m_p_powerup = NULL;
		}
	}
}

void c_game::draw_powerup(void)
{
	if (m_p_powerup != NULL)
		m_p_powerup->draw(m_buf);
}

void c_game::update_player(void)
{
	m_player->update();
	
	// Check to see if player is dead, if so it's game over!
	if (m_player->get_state() == PLS_DEAD)
		set_run_state(GRS_GAMEOVER);
}

bool c_game::rect_hit_test(float l1, float t1, float r1, float b1, float l2, float t2, float r2, float b2)
{
	bool f_hit = true;
	float temp;
	
	// Swap co-ordinates into a sensible order
	if (l1 > l2)
	{
		temp = l2;
		l2 = l1;
		l1 = temp;
	}
	if (r1 > r2)
	{
		temp = r2;
		r2 = r1;
		r1 = temp;
	}
	if (t1 > t2)
	{
		temp = t2;
		t2 = t1;
		t1 = temp;
	}
	if (b1 > b2)
	{
		temp = b2;
		b2 = b1;
		b1 = temp;
	}
	
	// Assume rectangles are intersecting, then find if rectangles are not
	// intersecting here (easier that checking if they are)
	if (b2 < t1)
		f_hit = false;
	if (t2 > b1)
		f_hit = false;
	if (r1 < l2)
		f_hit = false;
	if (l2 > r2)
		f_hit = false;
	
	return f_hit;
}

void c_game::add_particle_system(PARTICLE_SYSTEM_TYPE pst, float x, float y, float vec_dir_x, float vec_dir_y)
{
	particle_system_list *psl = new particle_system_list;
	
	if (psl == NULL)
		return;
	
	psl->ps = new c_particle_system(pst, x, y, vec_dir_x, vec_dir_y);
	if (psl->ps == NULL)
		return;
	
	// Add the new node to the list
	psl->p_next = m_p_particle_systems;
	m_p_particle_systems = psl;
}

void c_game::update_particle_systems(void)
{
	particle_system_list *psl = m_p_particle_systems, *p_prev = NULL, *p_temp = NULL;
	
	// Loop through all particle systems and update them
	while (psl != NULL)
	{
		psl->ps->update();
		
		if (!psl->ps->is_active())
		{
			// The particle system is no longer active
			
			// Draw the final state to the backdrop
			psl->ps->draw(m_backdrop);
			
			// remove the particle system from the list and destroy it
			p_temp = psl->p_next;
			if (p_prev != NULL)
				p_prev->p_next = p_temp;
			else
			{
				m_p_particle_systems = p_temp;
				p_prev = NULL;
			}
			delete psl->ps;
			delete psl;
			psl = p_temp;
		}
		else
		{
			p_prev = psl;
			psl = psl->p_next;
		}
	}
}

void c_game::draw_particle_systems(void)
{
	particle_system_list *psl = m_p_particle_systems;
	
	// Loop through the particle system list and draw them
	while (psl != NULL)
	{
		// If the type is PST_BLOOD, draw to the backdrop. This will give a
		// "smeared" effect. Otherwise draw to the buffer as normal.
		if (psl->ps->get_type() == PST_BLOOD)
			psl->ps->draw(m_backdrop);
		else
			psl->ps->draw(m_buf);
		psl = psl->p_next;
	}
}
	
int c_game::get_bound_top(void)
{
	return m_bound_top;
}

int c_game::get_bound_left(void)
{
	return m_bound_left;
}

int c_game::get_bound_bottom(void)
{
	return m_bound_bottom;
}

int c_game::get_bound_right(void)
{
	return m_bound_right;
}

void c_game::set_run_state(GAME_RUN_STATE state)
{
	// Set run state and play music accordingly
	switch (state)
	{
		case GRS_MAINMENU:
			m_p_jukebox->play_track(JBT_INTRO, false);
			break;
			
		case GRS_PLAYING:
			m_p_jukebox->play_track(m_curr_main_track, false);
			break;
			
		case GRS_PAUSE:
			break;
			
		case GRS_GAMEOVER:
			break;			
	}
	
	m_run_state = state;
}

bool c_game::reset(void)
{
	bool f_result = false;
	
	// Shutdown, then init the game with the reset flag set
	if (shutdown(true))
		if (init(true))
			f_result = true;
	
	return f_result;
}

bool c_game::init(bool f_reset)
{
	if (!f_reset && m_state != GS_NONE)
		return false;

	if (!f_reset)
	{
		// Seed the random number generator
		srand(time(NULL));
		
		// Set up the screen
		set_color_depth(32);
		if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0) != 0)
		{
			allegro_message("ERROR: Unable to initialise graphics -- aborting...", "");
			return false;
		}
		
		// Set the window title
		set_window_title("Malevolent Revenge");
	
		// Create the back buffer
		m_buf = create_bitmap(SCREEN_W, SCREEN_H);
		if (m_buf == NULL)
		{
			allegro_message("ERROR: Unable to initialise back buffer -- aborting...", "");
			return false;
		}
	}
	
	// Reload the backdrop if necessary
	if (m_backdrop != NULL)
	{
		destroy_bitmap(m_backdrop);
		m_backdrop = NULL;
	}
	
	// Load the game arena backdrop
	m_backdrop = load_bitmap("bitmaps/Backdrop.bmp", NULL);
	if (m_backdrop == NULL)
	{
		allegro_message("ERROR: Unable to load bitmaps/Backdrop.bmp -- aborting...", "");
		return false;
	}
		
	if (!f_reset)
	{
		// Load the game over screen background
		m_game_over_backdrop = load_bitmap("bitmaps/Game_over_screen.bmp", NULL);
		if (m_game_over_backdrop == NULL)
		{
			allegro_message("ERROR: Unable to load bitmaps/Game_over_screen.bmp -- aborting...", NULL);
			return false;
		}
		
		install_keyboard();
		install_mouse();
		
		// Load the custom mouse cursor, or default to the hardware cursor
		m_mouse_cursor = load_bitmap("bitmaps/Mouse_cursor.bmp", NULL);
		if (m_mouse_cursor != NULL)
		{
			set_mouse_sprite(m_mouse_cursor);
			set_mouse_sprite_focus(0, 0);
			select_mouse_cursor(MOUSE_CURSOR_ALLEGRO);
		}
		else
		{
			enable_hardware_cursor();
			select_mouse_cursor(MOUSE_CURSOR_ARROW);
		}
		show_mouse(screen);
		
		// If compiling for Linux, force ESD sound, otherwise autodetect it
#ifdef LINUX
		install_sound(DIGI_ESD, MIDI_NONE, NULL);
#else
		install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);
#endif
		
		// Set up the gameplay bounds
		m_bound_left   = 0;
		m_bound_right  = SCREEN_W;
		m_bound_top    = 0;
		m_bound_bottom = SCREEN_H;
		
		// Initialise the jukebox
		m_p_jukebox = new c_jukebox;
		if (m_p_jukebox == NULL)
		{
			allegro_message("ERROR: Unable to create jukebox -- aborting...", "");
			return false;
		}
		m_p_jukebox->load_track(JBT_INTRO,  "music/menu_mus.wav");
		m_p_jukebox->load_track(JBT_MAIN_1, "music/bg_mus.wav");
		m_p_jukebox->load_track(JBT_MAIN_2, "music/bg_mus2.wav");
		
		// Initialise the SFX object
		m_p_sfx = new c_sfx;
		if (m_p_sfx == NULL)
		{
			allegro_message("ERROR: Unable to initialise sound F/X -- aborting...", "");
			return false;
		}
		m_p_sfx->set_enabled(true);
		
		// Initialise the main menu
		m_p_menu_main = new c_menu(this);
		if (m_p_menu_main == NULL)
		{
			allegro_message("ERROR: Unable to initialise main menu -- aborting...", "");
			return false;
		}
		m_p_menu_main->load_actor_heading("data/Heading_anim.dat");
		m_p_menu_main->add_item("Play game");
		m_p_menu_main->add_item("Quit");
		m_p_menu_main->load_message("data/Main_menu_message.dat");
		m_p_menu_main->load_backdrop("bitmaps/Title_screen.bmp");
		
		// Create the player
		m_player = new c_player(this, "data/Marle.dat");
		if (m_player == NULL)
		{
			allegro_message("ERROR: Unable to initialise player -- aborting...", "");
			return false;
		}
		m_player->set_health(PLAYER_HEALTH_MAX);
		
		m_state = GS_INIT;
	}
	else
	{
		// Reset gameplay variables for a new game
		m_player->set_health(PLAYER_HEALTH_MAX);
		m_player->set_state(PLS_ALIVE);
		m_score     = 0;
		m_shots     = 0;
		m_shots_hit = 0;
	}

	return true;
}

bool c_game::shutdown(bool f_reset)
{
	enemy_list           *p_enode  = m_p_enemies,          *p_etemp  = NULL;
	projectile_list      *p_pnode  = m_p_projectiles,      *p_ptemp  = NULL;
	particle_system_list *p_psnode = m_p_particle_systems, *p_pstemp = NULL;

	if (!f_reset)
		m_state = GS_SHUTDOWN;
	
	// Destroy all enemies
	while (p_enode != NULL)
	{
		p_etemp = p_enode;
		p_enode = p_enode->p_next;
		delete p_etemp->enemy;
		delete p_etemp;
	}
	m_p_enemies = NULL;
	m_enemy_count = 0;

	// Destroy all projectiles
	while (p_pnode != NULL)
	{
		p_ptemp = p_pnode;
		p_pnode = p_pnode->p_next;
		delete p_ptemp->projectile;
		delete p_ptemp;
	}
	m_p_projectiles = NULL;

	// Destroy all particle systems
	while (p_psnode != NULL)
	{
		p_pstemp = p_psnode;
		p_psnode = p_psnode->p_next;
		delete p_pstemp->ps;
		delete p_pstemp;
	}
	m_p_particle_systems = NULL;
	
	// Destroy the powerup if applicable
	if (m_p_powerup != NULL)
	{
		delete m_p_powerup;
		m_p_powerup = NULL;
	}
	
	if (!f_reset)
	{
		if (m_p_sfx != NULL)
		{
			delete m_p_sfx;
			m_p_sfx = NULL;
		}
		
		if (m_p_jukebox != NULL)
		{
			delete m_p_jukebox;
			m_p_jukebox = NULL;
		}
		
		if (m_p_menu_main != NULL)
		{
			delete m_p_menu_main;
			m_p_menu_main = NULL;
		}
	
		if (m_player != NULL)
		{
			delete m_player;
			m_player = NULL;
		}
	
		remove_sound();
		
		show_mouse(NULL);
		remove_mouse();
		if (m_mouse_cursor != NULL)
		{
			destroy_bitmap(m_mouse_cursor);
			m_mouse_cursor = NULL;
		}
		
		clear_keybuf();
		
		if (m_game_over_backdrop != NULL)
		{
			destroy_bitmap(m_game_over_backdrop);
			m_game_over_backdrop = NULL;
		}
		
		if (m_backdrop != NULL)
		{
			destroy_bitmap(m_backdrop);
			m_backdrop = NULL;
		}
		
		if (m_buf != NULL)
		{
			destroy_bitmap(m_buf);
			m_buf = NULL;
		}
	}
	return true;
}

void c_game::run(void)
{
	// Platform-specific timing variables
#ifdef LINUX
	struct timeval tv;
	uint64_t usec_s, usec_e;
#else
	long int ticks_s, ticks_e;
#endif
	
	int spawn_chance = 0;
	
	// If game is not currently in the initialised state, it can't be run
	if (m_state != GS_INIT)
	{
		allegro_message("ERROR: The game cannot be run as it has not been initialised", "");
		return;
	}
	
	m_state = GS_RUN;
	
	set_run_state(GRS_MAINMENU);
	
	while (m_state == GS_RUN)
	{
		// Platform-specific timing code
#ifdef LINUX
		gettimeofday(&tv, NULL);
		usec_s = (tv.tv_sec * 1000000) + tv.tv_usec;
#else
		ticks_s = GetTickCount();
#endif
		
		switch (m_run_state)
		{
			case GRS_MAINMENU:
				m_p_menu_main->update();
				m_p_menu_main->draw(m_buf);
				break;
				
			case GRS_PLAYING:
				// See if a Diablos should be spawned on this frame
				spawn_chance = 50 + (m_enemy_count * 15);
				if (m_enemy_count == 0)
					spawn_chance = 5;
				if (rand() % spawn_chance == 0)
					add_enemy(m_bound_left + (rand() % (m_bound_right - m_bound_left)), m_bound_top + (rand() % (m_bound_bottom - m_bound_top)), 100.0, 10, "data/Diablos.dat");
				
				// See if a Hench should be spawned on this frame
				spawn_chance = 300 + (m_enemy_count * 100);
				if (rand() % spawn_chance == 0)
					add_enemy(m_bound_left + (rand() % (m_bound_right - m_bound_left)), m_bound_top + (rand() % (m_bound_bottom - m_bound_top)), 200.0, 30, "data/Hench.dat");
				
				// If the powerup is inactive, see if it should be spawned on this frame then spawn it if so
				if (m_p_powerup == NULL)
				{
					spawn_chance = 1000;
					if (rand() % spawn_chance == 0)
					{
						m_p_powerup = new c_powerup(this, "data/Health.dat", PT_HEALTH, 50.0, 200);
						if (m_p_powerup != NULL)
						{
							m_p_powerup->set_anim("IDLE");
							m_p_powerup->set_x(m_bound_left + (rand() % (m_bound_right  - m_bound_left)));
							m_p_powerup->set_y(m_bound_top  + (rand() % (m_bound_bottom - m_bound_top)));
						}
					}
				}
		
				// Update everything
				update_enemies();
				update_projectiles();
				update_particle_systems();
				update_powerup();
				update_player();
				
				// Clear the buffer, draw everything, the draw score etc
				blit(m_backdrop, m_buf, 0, 0, 0, 0, m_buf->w, m_buf->h);
				draw_powerup();
				draw_enemies();
				m_player->draw(m_buf);
				draw_projectiles();
				draw_particle_systems();
				textprintf_ex(m_buf, font, 10, 10, makecol(255, 255, 255), -1, "Health: %d%%, score: %d, accuracy: %d%%", (int)((m_player->get_health() / PLAYER_HEALTH_MAX) * 100.0), m_score, (m_shots > 0 ? (int)(((float)m_shots_hit / (float)m_shots) * 100.0) : 0));
				break;
				
			case GRS_PAUSE:
				// Dont update anything, but draw everything inclusing "GAME PAUSED"
				blit(m_backdrop, m_buf, 0, 0, 0, 0, m_buf->w, m_buf->h);
				draw_powerup();
				draw_enemies();
				m_player->draw(m_buf);
				draw_projectiles();
				draw_particle_systems();
				textprintf_ex(m_buf, font, 100, 100, makecol(255, 255, 255), -1, "GAME PAUSED");
				break;
				
			case GRS_GAMEOVER:
				// Draw the game over screen background and final scores etc
				blit(m_game_over_backdrop, m_buf, 0, 0, 0, 0, m_game_over_backdrop->w, m_game_over_backdrop->h);
				textprintf_ex(m_buf, font, 120, 180, makecol(255, 255, 255), -1, "Score:       %d",   m_score);
				textprintf_ex(m_buf, font, 120, 195, makecol(255, 255, 255), -1, "Shots fired: %d",   m_shots);
				textprintf_ex(m_buf, font, 120, 210, makecol(255, 255, 255), -1, "Accuracy:    %d%%", (m_shots > 0 ? (int)(((float)m_shots_hit / (float)m_shots) * 100.0) : 0));
				textprintf_centre_ex(m_buf, font, m_bound_left + ((m_bound_right - m_bound_left) / 2), m_bound_bottom - 30, makecol(255, 255, 255), -1, "-- Press ENTER to play again, or ESC to return to the menu --");
				break;
		}

		// Handle key presses
		keyboard_handler();
		
		m_frame++;

		// Blit the back buffer to the primary screen (hiding the mouse whilst performing the blit)
		scare_mouse();
		blit(m_buf, screen, 0, 0, 0, 0, m_buf->w, m_buf->h);
		unscare_mouse();
		
		// Platform-specific timing code to keep frame rate at 30fps
#ifdef LINUX
		do
		{
			// Keep sleeping for 5ms until the time since the last frame is >= 33.3ms
			sleep(5);
			gettimeofday(&tv, NULL);
			usec_e = (tv.tv_sec * 1000000) + tv.tv_usec;
		} while (usec_e < (usec_s + 33333));
#else
		do
		{
			// Keep sleeping for 1ms until the time since the last frame is >= 32ms
			sleep(1);
			ticks_e = GetTickCount();
		} while (ticks_e < (ticks_s + 32));
#endif
	}
}
