#include "c_player.h"
#include "c_projectile.h"
#include "c_game.h"
#include "c_sprite.h"
#include <math.h>

c_player::c_player()
{
	m_state = PLS_ALIVE;
	m_bullet_timer = 0;
}

c_player::c_player(c_game *game, string dat_file):c_character_actor(game, dat_file)
{
	m_state = PLS_ALIVE;
	m_bullet_timer = 0;
}

c_player::~c_player()
{
}

void c_player::set_state(PLAYER_STATE state)
{
	m_state = state;
	switch (state)
	{
		case PLS_ALIVE:
			set_anim("IDLE");
		case PLS_DYING:
			set_anim("DYIG");
	}
}

PLAYER_STATE c_player::get_state(void)
{
	return m_state;
}

void c_player::fire_projectile(float vx, float vy)
{
	int x = 0, y = 0;
	
	if (m_bullet_timer == 0 && m_state == PLS_ALIVE)
	{
		// Start projectile at player centre point
		x = m_x + ((float)m_sprite->get_tile_width() / 2.0);
		y = m_y + ((float)m_sprite->get_tile_height() / 2.0);
		
		// Use c_game to fire the projectile
		m_p_game->fire_projectile(this, DT_ENEMY, x, y, vx, vy, "data/Player_bullet.dat");
		
		// Set the timer to limit player firing rate
		m_bullet_timer = PLAYER_PROJECTILE_LIMITER;
	}
}

void c_player::update(void)
{
	float dx = 0.0, dy = 0.0, player_x = 0.0, player_y = 0.0, mag = 0.0;
	
	switch (m_state)
	{
		case PLS_ALIVE:
			// Set velocity vector based on cursor key states
			if (key[KEY_LEFT] && !key[KEY_RIGHT])
				set_xvel(-m_speed);
			else if (key[KEY_RIGHT] && !key[KEY_LEFT])
				set_xvel(m_speed);
			else
				set_xvel(0.0);
			if (key[KEY_UP] && !key[KEY_DOWN])
				set_yvel(-m_speed);
			else if (key[KEY_DOWN] && !key[KEY_UP])
				set_yvel(m_speed);
			else
				set_yvel(0.0);
				
			// Set appropriate animations based on velocity
			if (m_xvel > 0.0)
				set_anim("MVRT");
			else if (m_xvel < 0.0)
				set_anim("MVLT");
			else if (m_yvel > 0.0)
				set_anim("MVDN");
			else if (m_yvel < 0.0)
				set_anim("MVUP");
			else if (m_xvel == 0.0 && m_yvel == 0.0)
				set_anim("IDLE");
			
			if (mouse_needs_poll())
				poll_mouse();
			
			// Player can either use mouse or keyboard to fire, mouse takes priority
			if (mouse_b && 0x01) // LMB pressed
			{
				// Calculate vector between player centre and mouse cursor
				player_x = m_x + ((float)(get_sprite()->get_tile_width())  / 2.0);
				player_y = m_y + ((float)(get_sprite()->get_tile_height()) / 2.0);
				dx  = mouse_x - player_x;
				dy  = mouse_y - player_y;
				
				// Normalise vector
				mag = sqrt(dx * dx + dy * dy);		
				dx /= mag;
				dy /= mag;
				
				// Fire a projectile using the calculated velocity vector scaled by PLAYER_PROJECTILE_SPEED
				fire_projectile(dx * PLAYER_PROJECTILE_SPEED, dy * PLAYER_PROJECTILE_SPEED);
			}
			else
			{
				// Fire a projectile in the appropriate direction if W, A, S or D are pressed
				if (key[KEY_W])
					fire_projectile(0.0, -PLAYER_PROJECTILE_SPEED);
				if (key[KEY_A])
					fire_projectile(-PLAYER_PROJECTILE_SPEED, 0.0);
				if (key[KEY_S])
					fire_projectile(0.0, PLAYER_PROJECTILE_SPEED);
				if (key[KEY_D])
					fire_projectile(PLAYER_PROJECTILE_SPEED, 0.0);
			}
			
			// Update bullet timer for rate limiting
			if (m_bullet_timer > 0)
				m_bullet_timer--;
			
			// Recharge player's health
			if (m_health < PLAYER_HEALTH_MAX)
				m_health += PLAYER_HEALTH_RECHARGE;
			
			// Call base class update()
			c_actor::update();
			
			// Keep player within the game bounds
			if (m_x < m_p_game->get_bound_left())
				m_x = m_p_game->get_bound_left();
			if (m_x + m_sprite->get_tile_width() > m_p_game->get_bound_right())
				m_x = m_p_game->get_bound_right() - m_sprite->get_tile_width();
			if (m_y < m_p_game->get_bound_top())
				m_y = m_p_game->get_bound_top();
			if (m_y + m_sprite->get_tile_height() > m_p_game->get_bound_bottom())
				m_y = m_p_game->get_bound_bottom() - m_sprite->get_tile_height();
			break;
		
		case PLS_DYING:
			if (m_sprite->is_curr_anim_finished())
				set_state(PLS_DEAD);
			break;
		case PLS_DEAD:
			break;
	}
}
