#include "c_sprite.h"

c_sprite::c_sprite()
{
	_init();
}

c_sprite::c_sprite(c_game *game, string filename, int tile_w, int tile_h)
{
	_init();
	m_p_game = game;
	load_bmp(filename.c_str());
	set_tile_width(tile_w);
	set_tile_height(tile_h);
}

c_sprite::~c_sprite()
{
	anim_frame *p_frame = NULL, *p_temp = NULL;
	
	// Destroy all defined animation frames
	for (int i = 0; i < m_anim_count; i++)
	{
		p_frame = m_anims[i].p_head;
		while (p_frame != NULL)
		{
			p_temp = p_frame;
			p_frame = p_frame->p_next;
			delete p_temp;
		}
		m_anims[i].p_head = m_anims[i].p_tail = NULL;
	}
	
	// Don't destroy m_bmp as that's the game's bitmap_factory's job
	m_bmp = NULL;
}

void c_sprite::_init(void)
{
	m_p_game        = NULL;
	m_bmp           = NULL;
	m_tile_w        = 0;
	m_tile_h        = 0;
	m_p_curr_frame  = NULL;
	m_curr_frame_no = 0;
	m_curr_anim     = -1;
	m_anim_count    = 0;
}

void c_sprite::step_anim(void)
{
	anim_frame *p_prev_frame = NULL;
	
	// Only step the animation if the speed_counter == the speed.
	m_anims[m_curr_anim].speed_counter++;
	if (m_anims[m_curr_anim].speed_counter == m_anims[m_curr_anim].speed)
	{
		p_prev_frame = m_p_curr_frame;
		if (m_anims[m_curr_anim].f_forward)
		{
			m_curr_frame_no++;
			m_p_curr_frame = m_p_curr_frame->p_next;
			if (m_curr_frame_no == m_anims[m_curr_anim].frame_count)
			{
				m_anims[m_curr_anim].f_finished = true;
				switch (m_anims[m_curr_anim].anim_type)
				{
					case AT_LOOP:
						m_curr_frame_no = 0;
						m_p_curr_frame = m_anims[m_curr_anim].p_head;
						break;
					case AT_PINGPONG:
						m_anims[m_curr_anim].f_forward = false;
						m_curr_frame_no--;
						m_p_curr_frame = m_anims[m_curr_anim].p_tail;
						break;
					case AT_ONESHOT:
						m_p_curr_frame = p_prev_frame;
						m_curr_frame_no--;
						break;
				}
			}
		}
		else
		{
			m_curr_frame_no--;
			m_p_curr_frame = m_p_curr_frame->p_prev;
			if (m_curr_frame_no == 0)
			{
				m_anims[m_curr_anim].f_finished = true;
				switch (m_anims[m_curr_anim].anim_type)
				{
					case AT_LOOP:
						m_curr_frame_no = m_anims[m_curr_anim].frame_count;
						m_p_curr_frame = m_anims[m_curr_anim].p_tail;
						break;
					case AT_PINGPONG:
						m_anims[m_curr_anim].f_forward = true;
						m_curr_frame_no++;
						m_p_curr_frame = m_anims[m_curr_anim].p_head;
						break;
					case AT_ONESHOT:
						m_p_curr_frame = p_prev_frame;
						m_curr_frame_no++;
						break;
				}
			}
		}
		
		// Reset the speed_counter ready for next step_anim() call
		m_anims[m_curr_anim].speed_counter = 0;
	}
}

int c_sprite::get_tile_width(void)
{
	return m_tile_w;
}

int c_sprite::get_tile_height(void)
{
	return m_tile_h;
}

void c_sprite::set_tile_width(int tw)
{
	// Only set the tile width if it's a suitable value
	if (tw > 0 && tw <= m_bmp->w && (m_bmp->w % tw == 0))
		m_tile_w = tw;
}

void c_sprite::set_tile_height(int th)
{
	// Only set the tile height if it's a suitable value
	if (th > 0 && th <= m_bmp->h && (m_bmp->h % th == 0))
		m_tile_h = th;
}

bool c_sprite::load_bmp(string filename)
{
	// Use m_p_game's bitmap factory to load the sprite sheet
	return m_p_game->get_bitmap_factory()->get_bitmap(filename, &m_bmp);
}

void c_sprite::draw(BITMAP *bmp_dest, int x, int y)
{
	if (m_p_curr_frame == NULL)
		return;
	
	// Use masked_blit so that #ff00ff colours are not drawn (transparent)
	masked_blit(m_bmp, bmp_dest, m_p_curr_frame->x * m_tile_w, m_p_curr_frame->y * m_tile_h, x, y, m_tile_w, m_tile_h);
	
	// This frame has been drawn, so step the animation
	step_anim();
}

int c_sprite::add_anim(ANIM_TYPE anim_type, int total_frames, int speed)
{
	if (m_anim_count == 16)
		return -1;
	
	m_anim_count++;
	m_anims[m_anim_count - 1].frame_count   = total_frames;
	m_anims[m_anim_count - 1].anim_type     = anim_type;
	m_anims[m_anim_count - 1].speed         = speed;
	m_anims[m_anim_count - 1].speed_counter = 0;
	m_anims[m_anim_count - 1].f_forward     = true;
	m_anims[m_anim_count - 1].f_finished    = false;
	m_anims[m_anim_count - 1].p_head        = NULL;
	m_anims[m_anim_count - 1].p_tail        = NULL;
	
	return m_anim_count - 1;
}

bool c_sprite::add_anim_frame(int anim_id, int x, int y)
{
	anim_frame *p_frame = NULL, *p_prev = NULL;
	
	// Validate the arguments
	if (anim_id < 0 || anim_id >= m_anim_count)
		return false;
	if (x < 0)
		return false;
	if (y < 0)
		return false;
	if (x * m_tile_w >= m_bmp->w)
		return false;
	if (y * m_tile_h >= m_bmp->h)
		return false;
	
	p_frame = p_prev = m_anims[anim_id].p_head;
	
	// Set p_prev to the penultimate frame
	while (p_frame != NULL)
	{
		p_prev  = p_frame;
		p_frame = p_frame->p_next;
	}
		
	p_frame = new anim_frame;
	if (p_frame == NULL)
		return false;
	p_frame->x = x;
	p_frame->y = y;
	
	// If the animation's head is NULL, set this frame to be the head and tail as it's the only one (so far)
	if (m_anims[anim_id].p_head == NULL)
	{
		m_anims[anim_id].p_head = p_frame;
		m_anims[anim_id].p_tail = p_frame;
		p_frame->p_prev = NULL;
		p_frame->p_next = NULL;
	}
	else
	{
		p_frame->p_prev = p_prev;
		p_prev->p_next  = p_frame;
		p_frame->p_next = NULL;
		m_anims[anim_id].p_tail = p_frame;
	}
	
	return true;
}

bool c_sprite::set_anim(int anim_id)
{
	if (anim_id < 0 || anim_id >= m_anim_count)
		return false;
	
	// If the current animation is the same as the one being set, do nothing but return success
	if (anim_id == m_curr_anim)
		return true;
	
	m_curr_anim     = anim_id;
	m_curr_frame_no = 0;
	m_p_curr_frame  = m_anims[m_curr_anim].p_head;
	m_anims[m_curr_anim].f_finished = false;

	return true;
}

int c_sprite::get_anim(void)
{
	return m_curr_anim;
}

bool c_sprite::is_curr_anim_finished(void)
{
	return m_anims[m_curr_anim].f_finished;
}
